from abc import ABC, abstractmethod
from math import floor
from typing import Tuple, List, Iterable, Dict
import numpy as np

from autogluon.searcher.bayesopt.datatypes.common import Hyperparameter, \
    Candidate
from autogluon.searcher.bayesopt.datatypes.scaling import Scaling


# Epsilon margin to account for numerical errors
EPS = 1e-8


class HyperparameterRange(ABC):
    """
    One can argue that this class does a bit too much: both definition of
    ranges and things like normalization, for now we do not plan to make the
    latter configurable, so it should be okay having everything in one place to
    keeps things fairly simple
    """
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @abstractmethod
    def to_ndarray(self, hp: Hyperparameter) -> np.ndarray:
        pass

    @abstractmethod
    def from_ndarray(self, cand_ndarray: np.ndarray) -> Hyperparameter:
        pass

    def ndarray_size(self) -> int:
        return 1

    @abstractmethod
    def from_zero_one(self, normalized_value: float) -> Hyperparameter:
        """
        Used to generate a random hp, takes as input a random value between 0.0
        and 1.0, which should be generated uniformly at random. Returns a valid
        active hyperparameter.

        Normally is the same as from_ndarray, but in the warm start case it is
        not. Will be limited to active ranges.
        """
        # by default this is the same as from_ndarray
        pass

    def random_hp(self, random_state) -> Hyperparameter:
        return self.from_zero_one(random_state.uniform(0.0, 1.0))


def scale_from_zero_one(
        value, lower_bound: float, upper_bound: float, scaling: Scaling):
    assert 0.0 <= value <= 1.0, value
    if lower_bound == upper_bound:
        return lower_bound
    else:
        lower = scaling.to_internal(lower_bound)
        upper = scaling.to_internal(upper_bound)
        range = upper - lower
        assert range > 0, (lower, upper)
        internal_value = value * range + lower
        hp = scaling.from_internal(internal_value)
        # set value in case it is off due to numerical rounding
        if hp < lower_bound:
            hp = lower_bound
        if hp > upper_bound:
            hp = upper_bound
        return hp


class HyperparameterRangeContinuous(HyperparameterRange):
    def __init__(
            self, name: str, lower_bound: float, upper_bound: float,
            scaling: Scaling, active_lower_bound: float = None,
            active_upper_bound: float = None):
        """
        Real valued hyperparameter.

        :param name: unique name of the hyperparameter.
        :param lower_bound: inclusive lower bound on all the values that
            parameter can take.
        :param upper_bound: inclusive upper bound on all the values that
            parameter can take.
        :param scaling: determines how the values of the parameter are enumerated internally.
            The parameter value is expressed as parameter = scaling(internal), where internal
            is internal representation of parameter, whcih is a real value, normally in range
            [0, 1]. To optimize the parameter, the internal is varied, and the parameter to be
            tested is calculated from such internal representation. This allows to speed up
            on many machine learning problems, where the parameter being optimized adjusts
            modelling capacity of the model. In such situation, often holds that
            capacity = log(parameter). See e.g. VC dimension for finite set-family:
            https://en.wikipedia.org/wiki/Vapnik–Chervonenkis_dimension#Bounds
        :param active_lower_bound: changes lower bound of the range of the parameter; Is used
            within warmstart functionality.
        :param active_upper_bound: changes upper bound on the values of parameter. Used for
            warmstart.
        """
        super().__init__(name)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.active_lower_bound = lower_bound if active_lower_bound is None else active_lower_bound
        self.active_upper_bound = upper_bound if active_upper_bound is None else active_upper_bound

        assert self.lower_bound <= self.active_upper_bound <= self.upper_bound
        assert self.lower_bound <= self.active_lower_bound <= self.upper_bound

        self.scaling = scaling

    def to_ndarray(self, hp: float) -> np.ndarray:
        assert isinstance(hp, float), (hp, self)
        assert self.lower_bound <= hp <= self.upper_bound, (hp, self)
        # convert everything to internal scaling, and then normalize between zero and one
        lower = self.scaling.to_internal(self.lower_bound)
        upper = self.scaling.to_internal(self.upper_bound)
        if upper == lower:
            result = 0.0  # if the bounds are fixed for a dimension
        else:
            result = (self.scaling.to_internal(hp) - lower) / (upper - lower)
        assert 0.0 <= result <= 1.0, (result, self)
        return np.array([result])

    def from_ndarray(self, ndarray: np.ndarray) -> float:
        scalar = ndarray.item()
        return scale_from_zero_one(scalar, self.lower_bound,
                                   self.upper_bound, self.scaling)

    def __repr__(self) -> str:
        return "{}({}, {}, {}, {}, {}, {})".format(
            self.__class__.__name__, repr(self.name),
            repr(self.active_lower_bound), repr(self.active_upper_bound),
            repr(self.scaling), repr(self.lower_bound), repr(self.upper_bound)
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HyperparameterRangeContinuous):
            return self.name == other.name \
                   and self.lower_bound == other.lower_bound \
                   and self.upper_bound == other.upper_bound \
                   and self.scaling == other.scaling
        return False

    def from_zero_one(self, normalized_value: float) -> float:
        return scale_from_zero_one(normalized_value, self.active_lower_bound,
                                   self.active_upper_bound, self.scaling)


class HyperparameterRangeInteger(HyperparameterRange):
    def __init__(self, name: str, lower_bound: int, upper_bound: int, scaling: Scaling,
                 active_lower_bound: int = None, active_upper_bound: int = None):
        """
        Both bounds are INCLUDED in the valid values. Under the hood generates a continuous
        range from lower_bound - 0.5 to upper_bound + 0.5.
        See docs for continuous hyperparameter for more information.
        """
        super().__init__(name)
        # reduce the range by epsilon at both ends, to avoid corner cases where numerical rounding
        # would cause a value that would end up out of range by one
        active_lower_bound = lower_bound if active_lower_bound is None \
            else active_lower_bound
        active_upper_bound = upper_bound if active_upper_bound is None \
            else active_upper_bound

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.active_lower_bound = active_lower_bound
        self.active_upper_bound = active_upper_bound
        self.scaling = scaling

        self._continuous_range = HyperparameterRangeContinuous(
            name, lower_bound - 0.5 + EPS, upper_bound + 0.5 - EPS, scaling,
            active_lower_bound - 0.5 + EPS, active_upper_bound + 0.5 - EPS,
        )

    def to_ndarray(self, hp: int) -> np.ndarray:
        assert isinstance(hp, int), (hp, type(hp), self)
        return self._continuous_range.to_ndarray(float(hp))

    def from_ndarray(self, ndarray: np.ndarray) -> int:
        continuous = self._continuous_range.from_ndarray(ndarray)
        result = int(round(continuous))
        # just to be sure
        assert result >= self.lower_bound, (result, self)
        assert result <= self.upper_bound, (result, self)
        return result

    def __repr__(self) -> str:
        return "{}({}, {}, {}, {}, {}, {})".format(
            self.__class__.__name__, repr(self.name),
            repr(self.active_lower_bound), repr(self.active_upper_bound),
            repr(self.scaling), repr(self.lower_bound), repr(self.upper_bound)
        )

    def __eq__(self, other):
        if isinstance(other, HyperparameterRangeInteger):
            return self.name == other.name \
                   and self.lower_bound == other.lower_bound \
                   and self.upper_bound == other.upper_bound \
                   and self.active_lower_bound == other.active_lower_bound \
                   and self.active_upper_bound == other.active_upper_bound \
                   and self.scaling == other.scaling
        return False

    def from_zero_one(self, normalized_value: float) -> Hyperparameter:
        scaling_result = scale_from_zero_one(
            normalized_value, self._continuous_range.active_lower_bound,
            self._continuous_range.active_upper_bound, self.scaling)
        result = int(round(scaling_result))
        return result


class HyperparameterRangeCategorical(HyperparameterRange):
    def __init__(self, name: str, choices: Tuple[str, ...],
                 active_choices: Tuple[str, ...] = None):
        """
        Can take on discrete set of values.
        :param name: name of dimension.
        :param choices: possible values of the hyperparameter.
        :param active_choices: a subset of choices, restricts choices to some subset of values.
        """
        super().__init__(name)
        # sort the choices, so that we are sure the order is the same when we do one hot encoding
        choices = sorted(choices)
        self.choices = choices
        self.active_choices = choices if active_choices is None else active_choices
        self.active_choices = sorted(self.active_choices)

        assert set(self.choices).issuperset(self.active_choices)

    def to_ndarray(self, hp: str) -> np.ndarray:
        assert hp in self.choices, "{} not in {}".format(hp, self)
        idx = self.choices.index(hp)
        result = np.zeros(shape=(len(self.choices),))
        result[idx] = 1.0
        return result

    def from_ndarray(self, cand_ndarray: np.ndarray) -> str:
        assert len(cand_ndarray) == len(self.choices), (cand_ndarray, self)
        return self.choices[int(np.argmax(cand_ndarray))]

    def ndarray_size(self) -> int:
        return len(self.choices)

    def from_zero_one(self, normalized_value: float) -> str:
        if normalized_value == 1.0:
            return self.active_choices[-1]
        else:
            return self.active_choices[int(floor(normalized_value * len(self.active_choices)))]

    def __repr__(self) -> str:
        return "{}({}, {})".format(
            self.__class__.__name__, repr(self.name), repr(self.choices)
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, HyperparameterRangeCategorical):
            return self.name == other.name \
                   and self.choices == other.choices
        return False


class HyperparameterRanges(ABC):
    @abstractmethod
    def to_ndarray(self, cand_tuple: Candidate) -> np.ndarray:
        pass

    def to_ndarray_matrix(self, candidates: Iterable[Candidate]) -> np.ndarray:
        return np.vstack([self.to_ndarray(cand) for cand in candidates])

    @abstractmethod
    def ndarray_size(self) -> int:
        pass

    @abstractmethod
    def from_ndarray(self, cand_ndarray: np.ndarray) -> Candidate:
        """
        Converts a candidate from internal ndarray representation (fed to the
        GP) into an external Candidate. This typically involves rounding.

        For numerical HPs it assumes values scaled between 0.0 and 1.0, for
        categorical HPs it assumes one scalar per category, which will convert
        to the category with the highest value (think of it as a soft one hot
        encoding).
        """
        pass

    @abstractmethod
    def random_candidate(self, random_state) -> Candidate:
        pass

    @abstractmethod
    def random_candidates(self, random_state, num_configs: int) -> \
            List[Candidate]:
        pass

    @abstractmethod
    def get_ndarray_bounds(self) -> List[Tuple[float, float]]:
        """
        Returns (lower, upper) bounds for each dimension in ndarray vector
        representation.

        :return: [(lower0, upper0), (lower1, upper1), ...]
        """
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    # Needed because some code iterates over HyperparameterRange entries
    @property
    def hp_ranges(self) -> Tuple[HyperparameterRange]:
        raise NotImplementedError(
            "Only some HyperparameterRanges subclasses implement hp_ranges")


class HyperparameterRanges_Impl(HyperparameterRanges):
    """
    Alternative to HyperparameterRanges_CS, without depending on ConfigSpace.

    """
    def __init__(self, *args: HyperparameterRange):
        """
        Usage:

        hp_ranges = HyperparameterRanges_Impl(hp1_range, hp2_range, hp3_range)
        """
        self._hp_ranges = tuple(args)

        names = [hp_range.name for hp_range in args]
        if len(set(names)) != len(names):
            raise ValueError("duplicate names in {}".format(names))

    def to_ndarray(self, cand_tuple: Candidate) -> np.ndarray:
        pieces = [hp_range.to_ndarray(hp) for hp_range, hp in zip(self._hp_ranges, cand_tuple)]
        return np.hstack(pieces)

    def ndarray_size(self) -> int:
        return sum(d.ndarray_size() for d in self._hp_ranges)

    def from_ndarray(self, cand_ndarray: np.ndarray) -> Candidate:
        """
        Converts a candidate from internal ndarray representation (fed to the GP) into an
        external Candidate

        For numerical HPs it assumes values scaled between 0.0 and 1.0, for categorical HPs it
        assumes one scalar per category, which will convert to the category with the highest value
        (think of it as a soft one hot encoding).

        all such values should be concatenated, see to_ndarray for the opposite conversion
        """

        # check size
        expected_size = self.ndarray_size()
        assert cand_ndarray.shape == (expected_size,), (cand_ndarray.shape, expected_size)

        hps = []
        start = 0
        for hp_range in self._hp_ranges:
            end = start + hp_range.ndarray_size()
            hps.append(hp_range.from_ndarray(cand_ndarray[start:end]))
            start = end

        return tuple(hps)

    def get_ndarray_bounds(self) -> List[Tuple[float, float]]:
        # Context values are fixed by setting min = max in bounds of lbfgs minimizer:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.Bounds.html#scipy.optimize.Bounds
        bounds = []
        for hp_range in self._hp_ranges:
            if isinstance(hp_range, HyperparameterRangeCategorical):
                for category in hp_range.choices:
                    if category in hp_range.active_choices:
                        if len(hp_range.active_choices) == 1:
                            bounds.append((1.0, 1.0))  # constrain choice to the only one which is active
                        else:
                            bounds.append((0.0, 1.0))
                    else:
                        bounds.append((0.0, 0.0))
            else:
                # conversion below includes log / reverse log / etc scaling
                # due to warm start the below could be different from global
                # bounds
                low = hp_range.to_ndarray(hp_range.from_zero_one(0.0)).item()
                high = hp_range.to_ndarray(hp_range.from_zero_one(1.0)).item()
                bounds.append((low, high))
        return bounds

    def refine_ndarray_bounds(
            self, bounds: List[Tuple[float, float]], candidate: Candidate,
            margin: float) -> List[Tuple[float, float]]:
        new_bounds = []
        bound_it = iter(bounds)
        for i, hp_range in enumerate(self._hp_ranges):
            if isinstance(hp_range, HyperparameterRangeCategorical):
                for _ in range(len(hp_range.choices)):
                    new_bounds.append(next(bound_it))
            else:
                low, high = next(bound_it)
                x = hp_range.to_ndarray(candidate[i]).item()
                low = max(low, x - margin)
                high = min(high, x + margin)
                new_bounds.append((low, high))
        assert len(bounds) == len(new_bounds)  # Sanity check
        return new_bounds

    def to_kwargs(self, cand_tuple: Candidate) -> Dict[str, Hyperparameter]:
        return {hp_range.name: hp for hp_range, hp in zip(self._hp_ranges, cand_tuple)}

    def random_candidate(self, random_state) -> Candidate:
        return tuple(hp_range.random_hp(random_state) for hp_range in self._hp_ranges)

    def random_candidates(self, random_state, num_configs: int) -> \
            List[Candidate]:
        return [self.random_candidate(random_state)
                for _ in range(num_configs)]

    def __repr__(self) -> str:
        return "{}{}".format(
            self.__class__.__name__, repr(self._hp_ranges)
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, HyperparameterRanges_Impl):
            return self._hp_ranges == other.hp_ranges
        return False

    def __len__(self) -> int:
        return len(self._hp_ranges)

    @property
    def hp_ranges(self) -> Tuple[HyperparameterRange]:
        return self._hp_ranges



import logging
from collections import defaultdict

import numpy as np

logger = logging.getLogger(__name__)


def get_type_family(dtype):
    """From dtype, gets the dtype family."""
    try:
        if dtype.name is 'category':
            return 'category'
        if 'datetime' in dtype.name:
            return 'datetime'
        elif np.issubdtype(dtype, np.integer):
            return 'int'
        elif np.issubdtype(dtype, np.floating):
            return 'float'
    except Exception as err:
        logger.exception(f'Warning: dtype {dtype} is not recognized as a valid dtype by numpy! AutoGluon may incorrectly handle this feature...')
        logger.exception(err)

    if dtype.name in ['bool', 'bool_']:
        return 'bool'
    elif dtype.name in ['str', 'string', 'object']:
        return 'object'
    else:
        return dtype.name


def get_type_groups_df(df):
    features_types = df.dtypes.to_dict()

    features_type_groups = defaultdict(list)
    features_types_tmp = {k: v.name for k, v in features_types.items()}
    for key, val in features_types_tmp.items():
        features_type_groups[val].append(key)
    return features_type_groups


def get_type_family_groups_df(df):
    features_types = df.dtypes.to_dict()

    features_type_groups = defaultdict(list)
    features_types_tmp = {k: get_type_family(v) for k, v in features_types.items()}
    for key, val in features_types_tmp.items():
        features_type_groups[val].append(key)
    return features_type_groups



import copy
import logging
import argparse
import functools
from collections import OrderedDict
import numpy as np
import multiprocessing as mp
import ConfigSpace as CS

from .space import *
from .space import _add_hp, _add_cs, _rm_hp, _strip_config_space, SPLITTER
from ..utils import EasyDict as ezdict
from ..utils.deprecate import make_deprecate

__all__ = ['args', 'obj', 'func', 'sample_config',
           'autogluon_register_args', 'autogluon_object', 'autogluon_function',
           'autogluon_register_dict']

logger = logging.getLogger(__name__)


def sample_config(args, config):
    args = copy.deepcopy(args)
    striped_keys = [k.split(SPLITTER)[0] for k in config.keys()]
    if isinstance(args, (argparse.Namespace, argparse.ArgumentParser)):
        args_dict = vars(args)
    else:
        args_dict = args
    for k, v in args_dict.items():
        # handle different type of configurations
        if k in striped_keys:
            if isinstance(v, NestedSpace):
                sub_config = _strip_config_space(config, prefix=k)
                args_dict[k] = v.sample(**sub_config)
            else:
                if SPLITTER in k:
                    continue
                args_dict[k] = config[k]
        elif isinstance(v, AutoGluonObject):
            args_dict[k] = v.init()
    return args

class _autogluon_method(object):
    SEED = mp.Value('i', 0)
    LOCK = mp.Lock()
    def __init__(self, f):
        self.f = f
        self.args = ezdict()
        functools.update_wrapper(self, f)

    def __call__(self, args, config={}, **kwargs):
        new_config = copy.deepcopy(config)
        self._rand_seed()
        args = sample_config(args, new_config)
        from ..scheduler.reporter import FakeReporter
        if 'reporter' not in kwargs:
            logger.debug('Creating FakeReporter for test purpose.')
            kwargs['reporter'] = FakeReporter()

        output = self.f(args, **kwargs)
        logger.debug('Reporter Done!')
        kwargs['reporter'](done=True)
        return output
 
    def register_args(self, default={}, **kwvars):
        if isinstance(default, (argparse.Namespace, argparse.ArgumentParser)):
            default = vars(default)
        self.kwvars = {}
        self.args = ezdict()
        self.args.update(default)
        self.update(**kwvars)

    def update(self, **kwargs):
        """For searcher support ConfigSpace
        """
        self.kwvars.update(kwargs)
        for k, v in self.kwvars.items():
            if isinstance(v, (NestedSpace)):
                self.args.update({k: v})
            elif isinstance(v, Space):
                hp = v.get_hp(name=k)
                self.args.update({k: hp.default_value})
            else:
                self.args.update({k: v})

    @property
    def cs(self):
        cs = CS.ConfigurationSpace()
        for k, v in self.kwvars.items():
            if isinstance(v, NestedSpace):
                _add_cs(cs, v.cs, k)
            elif isinstance(v, Space):
                hp = v.get_hp(name=k)
                _add_hp(cs, hp)
            else:
                _rm_hp(cs, k)
        return cs

    @property
    def kwspaces(self):
        """For RL searcher/controller
        """
        kw_spaces = OrderedDict()
        for k, v in self.kwvars.items():
            if isinstance(v, NestedSpace):
                if isinstance(v, Categorical):
                    kw_spaces['{}{}choice'.format(k, SPLITTER)] = v
                for sub_k, sub_v in v.kwspaces.items():
                    new_k = '{}{}{}'.format(k, SPLITTER, sub_k)
                    kw_spaces[new_k] = sub_v
            elif isinstance(v, Space):
                kw_spaces[k] = v
        return kw_spaces

    def _rand_seed(self):
        _autogluon_method.SEED.value += 1
        np.random.seed(_autogluon_method.SEED.value)

    def __repr__(self):
        return repr(self.f)


def args(default=None, **kwvars):
    """Decorator for a Python training script that registers its arguments as hyperparameters. 
       Each hyperparameter takes fixed value or is a searchable space, and the arguments may either be:
       built-in Python objects (e.g. floats, strings, lists, etc.), AutoGluon objects (see :func:`autogluon.obj`), 
       or AutoGluon search spaces (see :class:`autogluon.space.Int`, :class:`autogluon.space.Real`, etc.).

    Examples
    --------
    >>> import autogluon as ag
    >>> @ag.args(batch_size=10, lr=ag.Real(0.01, 0.1))
    >>> def train_func(args):
    ...     print('Batch size is {}, LR is {}'.format(args.batch_size, arg.lr))
    """
    if default is None:
        default = dict()
    kwvars['_default_config'] = default
    def registered_func(func):
        @_autogluon_method
        @functools.wraps(func)
        def wrapper_call(*args, **kwargs):
            return func(*args, **kwargs)

        default = kwvars['_default_config']
        wrapper_call.register_args(default=default, **kwvars)
        return wrapper_call

    return registered_func


def func(**kwvars):
    """Decorator for a function that registers its arguments as hyperparameters. 
       Each hyperparameter may take a fixed value or be a searchable space (autogluon.space).

    Returns
    -------
    Instance of :class:`autogluon.space.AutoGluonObject`:
        A lazily initialized object, which allows for distributed training.

    Examples
    --------
    >>> import autogluon as ag
    >>> from gluoncv.model_zoo import get_model
    >>> 
    >>> @ag.func(pretrained=ag.space.Categorical(True, False))
    >>> def cifar_resnet(pretrained):
    ...     return get_model('cifar_resnet20_v1', pretrained=pretrained)
    """
    def _autogluon_kwargs_func(**kwvars):
        def registered_func(func):
            kwspaces = OrderedDict()
            @functools.wraps(func)
            def wrapper_call(*args, **kwargs):
                _kwvars = copy.deepcopy(kwvars)
                _kwvars.update(kwargs)
                for k, v in _kwvars.items():
                    if isinstance(v, NestedSpace):
                        kwspaces[k] = v
                        kwargs[k] = v
                    elif isinstance(v, Space):
                        kwspaces[k] = v
                        hp = v.get_hp(name=k)
                        kwargs[k] = hp.default_value
                    else:
                        kwargs[k] = v
                return func(*args, **kwargs)
            wrapper_call.kwspaces = kwspaces
            return wrapper_call
        return registered_func

    def registered_func(func):
        class autogluonobject(AutoGluonObject):
            @_autogluon_kwargs_func(**kwvars)
            def __init__(self, *args, **kwargs):
                self.func = func
                self.args = args
                self.kwargs = kwargs
                self._inited = False

            def sample(self, **config):
                kwargs = copy.deepcopy(self.kwargs)
                kwspaces = copy.deepcopy(autogluonobject.kwspaces)
                for k, v in kwargs.items():
                    if k in kwspaces and isinstance(kwspaces[k], NestedSpace):
                        sub_config = _strip_config_space(config, prefix=k)
                        kwargs[k] = kwspaces[k].sample(**sub_config)
                    elif k in config:
                        kwargs[k] = config[k]
                        
                return self.func(*self.args, **kwargs)

        @functools.wraps(func)
        def wrapper_call(*args, **kwargs):
            _kwvars = copy.deepcopy(kwvars)
            _kwvars.update(kwargs)
            agobj = autogluonobject(*args, **kwargs)
            agobj.kwvars = _kwvars
            return agobj
        return wrapper_call
    return registered_func

def obj(**kwvars):
    """Decorator for a Python class that registers its arguments as hyperparameters. 
       Each hyperparameter may take a fixed value or be a searchable space (autogluon.space).

    Returns
    -------
    Instance of :class:`autogluon.space.AutoGluonObject`:
        A lazily initialized object, which allows distributed training.

    Examples
    --------
    >>> import autogluon as ag
    >>> from mxnet import optimizer as optim
    >>> @ag.obj(
    >>>     learning_rate=ag.space.Real(1e-4, 1e-1, log=True),
    >>>     wd=ag.space.Real(1e-4, 1e-1),
    >>> )
    >>> class Adam(optim.Adam):
    >>>     pass
    """
    def _autogluon_kwargs_obj(**kwvars):
        def registered_func(func):
            kwspaces = OrderedDict()
            @functools.wraps(func)
            def wrapper_call(*args, **kwargs):
                kwvars.update(kwargs)
                for k, v in kwvars.items():
                    if isinstance(v, NestedSpace):
                        kwspaces[k] = v
                        kwargs[k] = v
                    elif isinstance(v, Space):
                        kwspaces[k] = v
                        hp = v.get_hp(name=k)
                        kwargs[k] = hp.default_value
                    else:
                        kwargs[k] = v
                return func(*args, **kwargs)
            wrapper_call.kwspaces = kwspaces
            wrapper_call.kwvars = kwvars
            return wrapper_call
        return registered_func

    def registered_class(Cls):
        class autogluonobject(AutoGluonObject):
            @_autogluon_kwargs_obj(**kwvars)
            def __init__(self, *args, **kwargs):
                self.args = args
                self.kwargs = kwargs
                self._inited = False

            def sample(self, **config):
                kwargs = copy.deepcopy(self.kwargs)
                kwspaces = copy.deepcopy(autogluonobject.kwspaces)
                for k, v in kwargs.items():
                    if k in kwspaces and isinstance(kwspaces[k], NestedSpace):
                        sub_config = _strip_config_space(config, prefix=k)
                        kwargs[k] = kwspaces[k].sample(**sub_config)
                    elif k in config:
                        kwargs[k] = config[k]

                args = self.args
                return Cls(*args, **kwargs)

            def __repr__(self):
                return 'AutoGluonObject -- ' + Cls.__name__

        autogluonobject.kwvars = autogluonobject.__init__.kwvars
        autogluonobject.__doc__ = Cls.__doc__
        autogluonobject.__name__ = Cls.__name__
        return autogluonobject

    return registered_class



autogluon_register_args = make_deprecate(args, 'autogluon_register_args')
autogluon_register_dict = make_deprecate(args, 'autogluon_register_dict')
autogluon_function = make_deprecate(func, 'autogluon_function')
autogluon_object = make_deprecate(obj, 'autogluon_object')



from abc import ABC, abstractmethod
import numpy as np
from scipy.special import gammaln
import numbers

from autogluon.searcher.bayesopt.gpmxnet.constants import MIN_POSTERIOR_VARIANCE

__all__ = ['Distribution',
           'Gamma',
           'Uniform',
           'Normal',
           'LogNormal',
           'Horseshoe']


class Distribution(ABC):
    @abstractmethod
    def negative_log_density(self, F, x):
        """
        Negative log density, computed in MXNet. lower and upper limits are
        ignored. If x is not a scalar, the distribution is i.i.d. over all
        entries.
        """
        pass


class Gamma(Distribution):
    """
    Gamma(mean, alpha):

        p(x) = C(alpha, beta) x^{alpha - 1} exp( -beta x), beta = alpha / mean,
        C(alpha, beta) = beta^alpha / Gamma(alpha)

    """
    def __init__(self, mean, alpha):
        self._assert_positive_number(mean, 'mean')
        self._assert_positive_number(alpha, 'alpha')
        self.mean = np.maximum(mean, MIN_POSTERIOR_VARIANCE)
        self.alpha = np.maximum(alpha, MIN_POSTERIOR_VARIANCE)
        self.beta = self.alpha / self.mean
        self.log_const = gammaln(self.alpha) - self.alpha * np.log(self.beta)
        self.__call__ = self.negative_log_density

    @staticmethod
    def _assert_positive_number(x, name):
        assert isinstance(x, numbers.Real) and x > 0.0, \
            "{} = {}, must be positive number".format(name, x)

    def negative_log_density(self, F, x):
        x_safe = F.maximum(x, MIN_POSTERIOR_VARIANCE)
        return F.sum(
            (1.0 - self.alpha) * F.log(x_safe) + self.beta * x_safe +
            self.log_const)

    def __call__(self, F, x):
        return self.negative_log_density(F, x)


class Uniform(Distribution):
    def __init__(self, lower: float, upper: float):
        self.log_const = np.log(upper - lower)
        self.__call__ = self.negative_log_density

    def negative_log_density(self, F, x):
        return F.sum(F.ones_like(x)) * self.log_const

    def __call__(self, F, x):
        return self.negative_log_density(F, x)


class Normal(Distribution):
    def __init__(self, mean: float, sigma: float):
        self.mean = mean
        self.sigma = sigma
        self.__call__ = self.negative_log_density

    def negative_log_density(self, F, x):
        return F.sum(F.square(x - self.mean)) * (0.5 / np.square(self.sigma))

    def __call__(self, F, x):
        return self.negative_log_density(F, x)


class LogNormal(Distribution):
    def __init__(self, mean: float, sigma: float):
        self.mean = mean
        self.sigma = sigma
        self.__call__ = self.negative_log_density

    def negative_log_density(self, F, x):
        x_safe = F.maximum(x, MIN_POSTERIOR_VARIANCE)
        return F.sum(
            F.log(x_safe * self.sigma) +
            F.square(F.log(x_safe) - self.mean) * (0.5 / np.square(self.sigma)))

    def __call__(self, F, x):
        return self.negative_log_density(F, x)


class Horseshoe(Distribution):
    def __init__(self, s: float):
        assert s > 0.0
        self.s = max(s, MIN_POSTERIOR_VARIANCE)
        self.__call__ = self.negative_log_density

    def negative_log_density(self, F, x):
        arg = F.maximum(3.0 * F.square(self.s / x), MIN_POSTERIOR_VARIANCE)
        return -F.sum(F.log(F.log1p(arg)))

    def __call__(self, F, x):
        return self.negative_log_density(F, x)



import logging
from collections import defaultdict

logger = logging.getLogger(__name__)


class FeatureTypesMetadata:
    """
    Contains feature type metadata information such as type family groups (feature_types_raw) and special feature type groups (feature_types_special)

    feature_types_raw is the dictionary computed as output to :function:`autogluon.utils.tabular.features.utils.get_type_family_groups_df`
    feature_types_special is an optional dictionary to communicate special properties of features to downstream models that have special handling functionality for those feature types.
        As an example, feature_types_special might contain a key 'text_ngram' indicating that the list of values are all features which were generated from a nlp vectorizer and represent ngrams.
        A downstream model such as a K-Nearest-Neighbor model could then check if 'text_ngram' is present in feature_types_special and drop those features if present, to speed up training and inference time.
    """
    def __init__(self, feature_types_raw: defaultdict, feature_types_special: defaultdict = None):
        # These dictionaries must have only 1 instance of a feature among all of its value lists.
        if feature_types_special is None:
            feature_types_special = defaultdict(list)

        self.feature_types_raw = feature_types_raw
        self.feature_types_special = feature_types_special

        self._validate()

    # Confirms if inputs are valid
    def _validate(self):
        feature_types_raw_expanded = []
        for key in self.feature_types_raw:
            feature_types_raw_expanded += self.feature_types_raw[key]
        feature_types_special_expanded = []
        for key in self.feature_types_special:
            feature_types_special_expanded += self.feature_types_special[key]
        feature_types_raw_expanded_set = set(feature_types_raw_expanded)
        if len(feature_types_raw_expanded) != len(feature_types_raw_expanded_set):
            raise AssertionError('feature_types_raw contains features that appear multiple times!')

        for feature in feature_types_special_expanded:
            if feature not in feature_types_raw_expanded:
                raise AssertionError(f"feature '{feature}' is present in feature_types_special but not in feature_types_raw!")

    def get_feature_type_raw(self, feature):
        return self._get_feature_type(feature=feature, feature_types_dict=self.feature_types_raw)

    def get_feature_type_special(self, feature):
        return self._get_feature_type(feature=feature, feature_types_dict=self.feature_types_special)

    def get_feature_types_raw_flattened(self):
        return {feature: type_family for type_family, features in self.feature_types_raw.items() for feature in features}

    @staticmethod
    def _get_feature_type(feature, feature_types_dict):
        for dtype_family in feature_types_dict:
            if feature in feature_types_dict[dtype_family]:
                return dtype_family
            else:
                raise ValueError(f'Feature {feature} not found in provided feature_types_dict!')



from typing import Callable, Optional, List
import mxnet as mx
import numpy as np
from mxnet import Context

from autogluon.searcher.bayesopt.gpmxnet.constants import DEFAULT_MCMC_CONFIG, \
    MCMCConfig
from autogluon.searcher.bayesopt.gpmxnet.gp_model import GaussianProcessModel
from autogluon.searcher.bayesopt.gpmxnet.kernel import KernelFunction
from autogluon.searcher.bayesopt.gpmxnet.likelihood import MarginalLikelihood
from autogluon.searcher.bayesopt.gpmxnet.mean import ScalarMeanFunction
from autogluon.searcher.bayesopt.gpmxnet.posterior_state import \
    GaussProcPosteriorState
from autogluon.searcher.bayesopt.gpmxnet.slice import SliceSampler
from autogluon.searcher.bayesopt.gpmxnet.gp_regression import \
    negative_log_posterior
from autogluon.searcher.bayesopt.gpmxnet.gluon_blocks_helpers import \
    encode_unwrap_parameter


class GPRegressionMCMC(GaussianProcessModel):

    def __init__(self,
                 build_kernel: Callable[[], KernelFunction],
                 mcmc_config: MCMCConfig = DEFAULT_MCMC_CONFIG,
                 random_seed: int = None,
                 ctx: Context = mx.cpu()):

        self.mcmc_config = mcmc_config

        if random_seed is not None:
            mx.random.seed(random_seed)
            self.random_seed = random_seed

        self.likelihood = _create_likelihood(build_kernel, ctx)
        self._states = None
        self.samples = None
        self.build_kernel = build_kernel
        self._ctx = ctx

    @property
    def states(self) -> Optional[List[GaussProcPosteriorState]]:
        return self._states

    @property
    def ctx(self):
        return self._ctx

    def fit(self, X, Y, **kwargs):
        F = mx.nd
        X = self._check_and_format_input(X)
        Y = self._check_and_format_input(Y)

        mean_function = self.likelihood.mean
        if isinstance(mean_function, ScalarMeanFunction):
            mean_function.set_mean_value(F.mean(Y).asscalar())

        def _log_posterior_density(hp_values: np.ndarray) -> float:
            # We check box constraints before converting hp_values to
            # internal
            if not self._is_feasible(hp_values):
                return -float('inf')
            # Decode and write into Gluon parameters
            _set_gp_hps(hp_values, self.likelihood)
            neg_log = negative_log_posterior(self.likelihood, X, Y)
            return -neg_log.asscalar()

        slice_sampler = SliceSampler(
            _log_posterior_density, 1.0, self.random_seed)
        init_hp_values = _get_gp_hps(self.likelihood)

        self.samples = slice_sampler.sample(
            init_hp_values, self.mcmc_config.n_samples,
            self.mcmc_config.n_burnin, self.mcmc_config.n_thinning)
        self._states = self._create_posterior_states(self.samples, X, Y)

    def recompute_states(self, X, Y, **kwargs):
        X = self._check_and_format_input(X)
        Y = self._check_and_format_input(Y)
        assert len(self.samples) > 0
        self._states = self._create_posterior_states(self.samples, X, Y)

    def _is_feasible(self, hp_values: np.ndarray) -> bool:
        pos = 0
        for _, encoding in self.likelihood.param_encoding_pairs():
            lower, upper = encoding.box_constraints()
            dim = encoding.dimension
            if lower is not None or upper is not None:
                values = hp_values[pos:(pos + dim)]
                if (lower is not None) and any(values < lower):
                    return False
                if (upper is not None) and any(values > upper):
                    return False
            pos += dim
        return True

    def _create_posterior_states(self, samples, X, Y):
        states = []
        for sample in samples:
            likelihood = _create_likelihood(self.build_kernel, self._ctx)
            _set_gp_hps(sample, likelihood)
            state = GaussProcPosteriorState(
                X, Y, likelihood.mean, likelihood.kernel,
                likelihood.get_noise_variance(as_ndarray=True))
            states.append(state)
        return states


def _get_gp_hps(likelihood: MarginalLikelihood) -> np.ndarray:
    """Get GP hyper-parameters as numpy array for a given likelihood object."""
    hp_values = []
    for param_int, encoding in likelihood.param_encoding_pairs():
        hp_values.append(
            encode_unwrap_parameter(mx.nd, param_int, encoding).asnumpy())
    return np.concatenate(hp_values)


def _set_gp_hps(params_numpy: np.ndarray, likelihood: MarginalLikelihood):
    """Set GP hyper-parameters from numpy array for a given likelihood object."""
    pos = 0
    for param, encoding in likelihood.param_encoding_pairs():
        dim = encoding.dimension
        values = params_numpy[pos:(pos + dim)]
        if dim == 1:
            internal_values = encoding.decode(values, param.name)
        else:
            internal_values = np.array(
                [encoding.decode(v, param.name) for v in values])
        param.set_data(internal_values)
        pos += dim


def _create_likelihood(build_kernel, ctx) -> MarginalLikelihood:
    """Create a MarginalLikelihood object with default initial GP hyperparameters."""

    likelihood = MarginalLikelihood(
        kernel=build_kernel(),
        mean=ScalarMeanFunction(),
        initial_noise_variance=None
    )
    likelihood.initialize(ctx=ctx, force_reinit=True)
    likelihood.hybridize()

    return likelihood



from typing import Iterator, Set, List
import numpy as np
import logging

from autogluon.searcher.bayesopt.datatypes.common import Candidate
from autogluon.searcher.bayesopt.datatypes.tuning_job_state import \
    TuningJobState
from autogluon.searcher.bayesopt.tuning_algorithms.base_classes import \
    CandidateGenerator
from autogluon.searcher.bayesopt.datatypes.hp_ranges import \
    HyperparameterRanges

logger = logging.getLogger(__name__)


class RandomCandidateGenerator(CandidateGenerator):
    def __init__(self, hp_ranges: HyperparameterRanges, random_seed: int):
        self.hp_ranges = hp_ranges
        self.random_seed = random_seed

    def generate_candidates(self) -> Iterator[Candidate]:
        random_state = np.random.RandomState(self.random_seed)
        while True:
            yield self.hp_ranges.random_candidate(random_state)

    def generate_candidates_en_bulk(self, num_cands: int) -> List[Candidate]:
        random_state = np.random.RandomState(self.random_seed)
        return self.hp_ranges.random_candidates(random_state, num_cands)


class RandomStatefulCandidateGenerator(CandidateGenerator):
    """
    As opposed to RandomCandidateGenerator, this generator maintains a
    random state, so if generate_candidates is called several times, different
    sequences are returned.

    """
    def __init__(self, hp_ranges: HyperparameterRanges,
                 random_state: np.random.RandomState):
        self.hp_ranges = hp_ranges
        self.random_state = random_state

    def generate_candidates(self) -> Iterator[Candidate]:
        while True:
            yield self.hp_ranges.random_candidate(self.random_state)

    def generate_candidates_en_bulk(self, num_cands: int) -> List[Candidate]:
        return self.hp_ranges.random_candidates(self.random_state, num_cands)


def compute_blacklisted_candidates(state: TuningJobState) -> Set[Candidate]:
    return set([x.candidate for x in state.candidate_evaluations] + \
               state.pending_candidates + state.failed_candidates)


MAX_RETRIES_ON_DUPLICATES = 10000


# ATTENTION: If this is used with Candidate = CS.Configuration, the overhead
# for filtering out duplicates and blacklisted configs becomes large
def generate_unique_candidates(
        candidates_generator: CandidateGenerator, num_candidates: int,
        blacklisted_candidates: Set[Candidate]) -> List[Candidate]:
    blacklisted = set(blacklisted_candidates)  # copy
    result = []
    num_results = 0
    retries = 0
    for i, cand in enumerate(candidates_generator.generate_candidates()):
        if cand not in blacklisted:
            result.append(cand)
            num_results += 1
            blacklisted.add(cand)
            retries = 0
        else:
            # found a duplicate; retry
            retries += 1

        # End loop if enough candidates where generated, or after too many retries
        # (this latter can happen when most of them are duplicates, and must be done
        # to avoid infinite loops in the purely discrete case)
        if num_results == num_candidates or retries > MAX_RETRIES_ON_DUPLICATES:
            if retries > MAX_RETRIES_ON_DUPLICATES:
                logger.warning(
                    f"Reached limit of {MAX_RETRIES_ON_DUPLICATES} retries with i={i}. "
                    f"Returning {len(result)} candidates instead of the requested {num_candidates}"
                )
            break

    return result



import copy
import logging
import math
import re
from collections import defaultdict

import numpy as np
import pandas as pd
import psutil
from pandas import DataFrame, Series
from pandas.api.types import CategoricalDtype

from .feature_types_metadata import FeatureTypesMetadata
from .utils import get_type_family, get_type_groups_df, get_type_family_groups_df
from ..utils.decorators import calculate_time
from ..utils.savers import save_pkl

logger = logging.getLogger(__name__)


# TODO: Add feature of # of observation counts to high cardinality categorical features
# TODO: Use code from problem type detection for column types! Ints/Floats could be Categorical through this method! Maybe try both?
class AbstractFeatureGenerator:
    def __init__(self):
        self.features_init_to_keep = []  # Original features to use as input to feature generation
        self.features_to_remove = []  # Original features to remove prior to feature generation
        self.features_to_remove_post = []  # Final features to remove prior to sending to models
        self.features_init_types = dict()  # Initial feature types prior to transformation
        self.feature_type_family_init_raw = defaultdict(list)  # Feature types of original features, without inferring.
        self.feature_type_family = defaultdict(list)  # Feature types of original features, after inferring.
        self.feature_type_family_generated = defaultdict(list)  # Feature types (special) of generated features.
        self.feature_transformations = defaultdict(list)  # Dictionary of transformation pipelines (keys) and the list of original features transformed through them (values)
        self.features_categorical_final = []  # Categorical features
        self.features_categorical_final_mapping = defaultdict()  # Categorical features original value -> category code mapping
        self.features_binned = []  # Features to be binned
        self.features_binned_mapping = defaultdict()  # Binned features original value ranges -> binned category code mapping
        self.features = []  # Final list of features after transformation

        self.fit = False  # Whether the feature generation has been fit

        self.minimize_categorical_memory_usage_flag = True

        self.pre_memory_usage = None
        self.pre_memory_usage_per_row = None
        self.post_memory_usage = None
        self.post_memory_usage_per_row = None
        self.is_dummy = False  # If True, returns a single dummy feature as output. Occurs if fit with no useful features.

        self.feature_types_metadata: FeatureTypesMetadata = None  # FeatureTypesMetadata object based on the final features. Passed to models to enable advanced functionality.

    def preprocess(self, X: DataFrame):
        return X

    # TODO: Save this to disk and remove from memory if large categoricals!
    @calculate_time
    def fit_transform(self, X: DataFrame, y=None, banned_features=None, fix_categoricals=False, drop_duplicates=True):
        self.fit = False
        X_len = len(X)
        self.pre_memory_usage = self.get_approximate_df_mem_usage(X, sample_ratio=0.2).sum()
        self.pre_memory_usage_per_row = self.pre_memory_usage / X_len
        available_mem = psutil.virtual_memory().available
        pre_memory_usage_percent = self.pre_memory_usage / (available_mem + self.pre_memory_usage)
        if pre_memory_usage_percent > 0.05:
            logger.warning(f'Warning: Data size prior to feature transformation consumes {round(pre_memory_usage_percent*100, 1)}% of available memory. Consider increasing memory or subsampling the data to avoid instability.')

        X_index = copy.deepcopy(X.index)
        X = X.reset_index(drop=True)  # TODO: Theoretically inplace=True avoids data copy, but can lead to altering of original DataFrame outside of method context.

        X_features = self._fit_transform(X=X, y=y, banned_features=banned_features, fix_categoricals=fix_categoricals, drop_duplicates=drop_duplicates)

        X_features.index = X_index

        for feature_type in self.feature_type_family_generated:
            self.feature_type_family_generated[feature_type] = list(set(self.feature_type_family_generated[feature_type]) - set(self.features_to_remove_post))

        self.post_memory_usage = self.get_approximate_df_mem_usage(X_features, sample_ratio=0.2).sum()
        self.post_memory_usage_per_row = self.post_memory_usage / X_len

        available_mem = psutil.virtual_memory().available
        post_memory_usage_percent = self.post_memory_usage / (available_mem + self.post_memory_usage + self.pre_memory_usage)

        if post_memory_usage_percent > 0.15:
            logger.warning(f'Warning: Data size post feature transformation consumes {round(post_memory_usage_percent*100, 1)}% of available memory. Consider increasing memory or subsampling the data to avoid instability.')

        if len(list(X_features.columns)) == 0:
            self.is_dummy = True
            logger.warning(f'WARNING: No useful features were detected in the data! AutoGluon will train using 0 features, and will always predict the same value. Ensure that you are passing the correct data to AutoGluon!')
            X_features['__dummy__'] = 0
            self.feature_type_family_generated['int'] = ['__dummy__']

        feature_types_raw = get_type_family_groups_df(X_features)
        self.feature_types_metadata = FeatureTypesMetadata(feature_types_raw=feature_types_raw, feature_types_special=self.feature_type_family_generated)

        self.features = list(X_features.columns)
        self.fit = True
        logger.log(20, 'Feature Generator processed %s data points with %s features' % (X_len, len(self.features)))
        self.print_feature_type_info()

        return X_features

    def _fit_transform(self, X: DataFrame, y=None, banned_features=None, fix_categoricals=False, drop_duplicates=True):
        if banned_features:
            self.features_to_remove += [str(feature) for feature in banned_features]

        X.columns = X.columns.astype(str)  # Ensure all column names are strings
        self.get_feature_types(X)
        X = X.drop(self.features_to_remove, axis=1, errors='ignore')
        self.features_init_to_keep = copy.deepcopy(list(X.columns))
        self.features_init_types = X.dtypes.to_dict()
        self.feature_type_family_init_raw = get_type_groups_df(X)

        X_features = self.generate_features(X)
        X_len = len(X)
        for column in X_features:
            unique_value_count = len(X_features[column].unique())
            if unique_value_count == 1:
                self.features_to_remove_post.append(column)
            # TODO: Consider making 0.99 a parameter to FeatureGenerator
            elif 'object' in self.feature_type_family and column in self.feature_type_family['object'] and (unique_value_count / X_len > 0.99):
                self.features_to_remove_post.append(column)

        self.features_binned = list(set(self.features_binned) - set(self.features_to_remove_post))
        self.features_binned_mapping = self.generate_bins(X_features, self.features_binned)
        for column in self.features_binned:  # TODO: Should binned columns be continuous or categorical if they were initially continuous? (Currently categorical)
            X_features[column] = self.bin_column(series=X_features[column], mapping=self.features_binned_mapping[column])
        X_features = X_features.drop(self.features_to_remove_post, axis=1)
        if drop_duplicates:
            X_features = self.drop_duplicate_features(X_features)
        self.features_categorical_final = list(X_features.select_dtypes(include='category').columns.values)
        if fix_categoricals:  # if X_test is not used in fit_transform and the model used is from SKLearn
            X_features = self.fix_categoricals_for_sklearn(X_features=X_features)
        for column in self.features_categorical_final:
            self.features_categorical_final_mapping[column] = copy.deepcopy(X_features[column].cat.categories)  # dict(enumerate(X_features[column].cat.categories))

        X_features = self.minimize_memory_usage(X_features=X_features)

        return X_features

    @calculate_time
    def transform(self, X: DataFrame):
        if not self.fit:
            raise AssertionError('FeatureGenerator has not yet been fit.')
        if self.features is None:
            raise AssertionError('FeatureGenerator.features is None, have you called fit() yet?')
        X_index = copy.deepcopy(X.index)
        X = X.reset_index(drop=True)
        X.columns = X.columns.astype(str)  # Ensure all column names are strings
        X = X.drop(self.features_to_remove, axis=1, errors='ignore')
        X_columns = X.columns.tolist()
        # Create any columns present in the training dataset that are now missing from this dataframe:
        missing_cols = []
        for col in self.features_init_to_keep:
            if col not in X_columns:
                missing_cols.append(col)
        if len(missing_cols) > 0:
            raise ValueError(f'Required columns are missing from the provided dataset. Missing columns: {missing_cols}')

        X = X.astype(self.features_init_types)
        X_features = self.generate_features(X)
        for column in self.features_binned:
            X_features[column] = self.bin_column(series=X_features[column], mapping=self.features_binned_mapping[column])
        if self.is_dummy:
            X_features.index = X_index
            X_features['__dummy__'] = 0
            return X_features
        X_features = X_features[self.features]
        for column in self.features_categorical_final:
            X_features[column].cat.set_categories(self.features_categorical_final_mapping[column], inplace=True)
        X_features = self.minimize_memory_usage(X_features=X_features)
        X_features.index = X_index

        return X_features

    def get_feature_types(self, X: DataFrame):
        features_init = list(X.columns)
        features_init = [feature for feature in features_init if feature not in self.features_to_remove]
        for column in features_init:
            mark_for_removal = False
            col_val = X[column]
            dtype = col_val.dtype
            num_unique = len(col_val.unique())

            type_family = self.get_type_family(dtype)
            if self.check_if_datetime_feature(col_val):
                type_family = 'datetime'
            elif self.check_if_nlp_feature(col_val):
                type_family = 'text'

            if num_unique == 1:
                mark_for_removal = True
            if mark_for_removal:
                self.features_to_remove.append(column)
            else:
                self.feature_type_family[type_family].append(column)

    def generate_features(self, X: DataFrame):
        raise NotImplementedError()

    # TODO: Expand to int64 -> date features (milli from epoch etc)
    def check_if_datetime_feature(self, X: Series):
        type_family = self.get_type_family(X.dtype)
        # TODO: Check if low numeric numbers, could be categorical encoding!
        # TODO: If low numeric, potentially it is just numeric instead of date
        if X.isnull().all():
            return False
        if type_family == 'datetime':
            return True
        if type_family != 'object':  # TODO: seconds from epoch support
            return False
        try:
            X.apply(pd.to_datetime)
            return True
        except:
            return False

    def check_if_nlp_feature(self, X: Series):
        type_family = self.get_type_family(X.dtype)
        if type_family != 'object':
            return False
        X_unique = X.unique()
        num_unique = len(X_unique)
        num_rows = len(X)
        unique_ratio = num_unique / num_rows
        # print(X_unique)
        if unique_ratio <= 0.01:
            return False
        avg_words = np.mean([len(re.sub(' +', ' ', value).split(' ')) if isinstance(value, str) else 0 for value in X_unique])
        # print(avg_words)
        if avg_words < 3:
            return False

        return True

    def minimize_memory_usage(self, X_features):
        if self.minimize_categorical_memory_usage_flag:
            X_features = self.minimize_categorical_memory_usage(X_features=X_features)
        X_features = self.minimize_ngram_memory_usage(X_features=X_features)
        X_features = self.minimize_binned_memory_usage(X_features=X_features)
        return X_features

    def minimize_ngram_memory_usage(self, X_features):
        if 'text_ngram' in self.feature_type_family_generated and self.feature_type_family_generated['text_ngram']:
            X_features[self.feature_type_family_generated['text_ngram']] = np.clip(X_features[self.feature_type_family_generated['text_ngram']], 0, 255).astype('uint8')
        return X_features

    def minimize_binned_memory_usage(self, X_features):
        if self.features_binned:
            X_features[self.features_binned] = np.clip(X_features[self.features_binned], 0, 255).astype('uint8')
        return X_features

    # TODO: Compress further, uint16, etc.
    # Warning: Performs in-place updates
    def minimize_categorical_memory_usage(self, X_features):
        cat_columns = X_features.select_dtypes('category').columns
        for column in cat_columns:
            new_categories = list(range(len(X_features[column].cat.categories.values)))
            X_features[column].cat.rename_categories(new_categories, inplace=True)
        return X_features

    def fix_categoricals_for_sklearn(self, X_features):
        for column in self.features_categorical_final:
            rank = X_features[column].value_counts().sort_values(ascending=True)
            rank = rank[rank >= 3]
            rank = rank.reset_index()
            val_list = list(rank['index'].values)
            if len(val_list) <= 1:
                self.features_to_remove_post.append(column)
                self.features_categorical_final = [feature for feature in self.features_categorical_final if feature != column]
                logger.debug(f'Dropping {column}')
            else:
                X_features[column] = X_features[column].astype(CategoricalDtype(categories=val_list))
        return X_features

    # TODO: add option for user to specify dtypes on load
    @staticmethod
    def get_type_family(dtype):
        return get_type_family(dtype=dtype)

    # TODO: optimize by not considering columns with unique sums/means
    # TODO: Multithread?
    @staticmethod
    def drop_duplicate_features(X):
        X_without_dups = X.T.drop_duplicates().T
        logger.debug(f"X_without_dups.shape: {X_without_dups.shape}")

        columns_orig = X.columns.values
        columns_new = X_without_dups.columns.values
        columns_removed = [column for column in columns_orig if column not in columns_new]

        del X_without_dups

        logger.log(15, 'Warning: duplicate columns removed ')
        logger.log(15, columns_removed)
        logger.log(15, f'Removed {len(columns_removed)} duplicate columns before training models')

        return X[columns_new]

    def get_feature_types_metadata_full(self):
        feature_types_metadata_full = copy.deepcopy(self.feature_types_metadata.feature_types_special)

        for key_raw in self.feature_types_metadata.feature_types_raw:
            values = self.feature_types_metadata.feature_types_raw[key_raw]
            for key_special in self.feature_types_metadata.feature_types_special:
                values = [value for value in values if value not in self.feature_types_metadata.feature_types_special[key_special]]
            if values:
                feature_types_metadata_full[key_raw] += values

        return feature_types_metadata_full

    @staticmethod
    def bin_column(series, mapping):
        mapping_dict = {k: v for v, k in enumerate(list(mapping))}
        series_out = pd.cut(series, mapping)
        # series_out.cat.categories = [str(g) for g in series_out.cat.categories]  # LightGBM crashes at end of training without this
        series_out_int = [mapping_dict[val] for val in series_out]
        return series_out_int

    # TODO: Rewrite with normalized value counts as binning technique, will be more performant and optimal
    @staticmethod
    def generate_bins(X_features: DataFrame, features_to_bin):
        X_len = len(X_features)
        ideal_cats = 10
        starting_cats = 1000
        bin_index_starting = [np.floor(X_len * (num + 1) / starting_cats) for num in range(starting_cats - 1)]
        bin_epsilon = 0.000000001
        bin_mapping = defaultdict()
        max_iterations = 20
        for column in features_to_bin:
            num_cats_initial = starting_cats
            bins_value_counts = X_features[column].value_counts(ascending=False, normalize=True)
            max_bins = len(bins_value_counts)

            if max_bins <= ideal_cats:
                bins = pd.Series(data=sorted(X_features[column].unique()))
                num_cats_initial = max_bins
                cur_len = max_bins
                bin_index = range(num_cats_initial)
                interval_index = AbstractFeatureGenerator.get_bins(bins=bins, bin_index=bin_index, bin_epsilon=bin_epsilon)
            else:
                cur_len = X_len
                bins = X_features[column].sort_values(ascending=True)
                interval_index = AbstractFeatureGenerator.get_bins(bins=bins, bin_index=bin_index_starting, bin_epsilon=bin_epsilon)

            max_desired_bins = min(ideal_cats, max_bins)
            min_desired_bins = min(ideal_cats, max_bins)

            if (len(interval_index) >= min_desired_bins) and (len(interval_index) <= max_desired_bins):
                is_satisfied = True
            else:
                is_satisfied = False

            num_cats_current = num_cats_initial
            # print(column, min_desired_bins, max_desired_bins)
            cur_iteration = 0
            while not is_satisfied:
                if len(interval_index) > max_desired_bins:
                    pass
                elif len(interval_index) < min_desired_bins:
                    pass
                ratio_reduction = max_desired_bins / len(interval_index)
                num_cats_current = int(np.floor(num_cats_current * ratio_reduction))
                bin_index = [np.floor(cur_len * (num + 1) / num_cats_current) for num in range(num_cats_current - 1)]
                interval_index = AbstractFeatureGenerator.get_bins(bins=bins, bin_index=bin_index, bin_epsilon=bin_epsilon)

                if (len(interval_index) >= min_desired_bins) and (len(interval_index) <= max_desired_bins):
                    is_satisfied = True
                    # print('satisfied', column, len(interval_index))
                cur_iteration += 1
                if cur_iteration >= max_iterations:
                    is_satisfied = True
                    # print('max_iterations met, stopping prior to satisfaction!', column, len(interval_index))

            bin_mapping[column] = interval_index
        return bin_mapping

    # TODO: Move this outside of here
    # TODO: Not accurate for categoricals, will count categorical mapping dict as taking more memory than it actually does.
    @staticmethod
    def get_approximate_df_mem_usage(df, sample_ratio=0.2):
        if sample_ratio >= 1:
            return df.memory_usage(deep=True)
        else:
            num_rows = len(df)
            num_rows_sample = math.ceil(sample_ratio * num_rows)
            return df.head(num_rows_sample).memory_usage(deep=True) / sample_ratio

    # TODO: Clean code
    # bins is a sorted int/float series, ascending=True
    @staticmethod
    def get_bins(bins: Series, bin_index, bin_epsilon):
        max_val = bins.max()
        bins_2 = bins.iloc[bin_index]
        bins_3 = list(bins_2.values)
        bins_unique = sorted(list(set(bins_3)))
        bins_with_epsilon_max = set([i for i in bins_unique] + [i - bin_epsilon for i in bins_unique if i == max_val])
        removal_bins = set([bins_unique[index - 1] for index, i in enumerate(bins_unique[1:], start=1) if i == max_val])
        bins_4 = sorted(list(bins_with_epsilon_max - removal_bins))
        bins_5 = [np.inf if (x == max_val) else x for x in bins_4]
        bins_6 = sorted(list(set([-np.inf] + bins_5 + [np.inf])))
        bins_7 = [(bins_6[i], bins_6[i + 1]) for i in range(len(bins_6) - 1)]
        interval_index = pd.IntervalIndex.from_tuples(bins_7)
        return interval_index

    def print_feature_type_info(self):
        logger.log(20, 'Original Features (raw dtypes):')
        for key, val in self.feature_type_family_init_raw.items():
            if val: logger.log(20, '\t%s features: %s' % (key, len(val)))
        logger.log(20, 'Original Features (inferred dtypes):')
        for key, val in self.feature_type_family.items():
            if val: logger.log(20, '\t%s features: %s' % (key, len(val)))
        logger.log(20, 'Generated Features (special dtypes):')
        for key, val in self.feature_types_metadata.feature_types_special.items():
            if val: logger.log(20, '\t%s features: %s' % (key, len(val)))
        logger.log(20, 'Processed Features (raw dtypes):')
        for key, val in self.feature_types_metadata.feature_types_raw.items():
            if val: logger.log(20, '\t%s features: %s' % (key, len(val)))
        logger.log(20, 'Processed Features:')
        for key, val in self.get_feature_types_metadata_full().items():
            if val: logger.log(20, '\t%s features: %s' % (key, len(val)))

    def save_self(self, path):
        save_pkl.save(path=path, object=self)



import numpy as np
import mxnet as mx
from mxnet.gluon import nn, HybridBlock
from mxnet.util import use_np
from autogluon_contrib_nlp.utils.config import CfgNode
from autogluon_contrib_nlp.layers import get_activation, get_norm_layer
from .. import constants as _C


@use_np
class BasicMLP(HybridBlock):
    def __init__(self, in_units,
                 mid_units,
                 out_units,
                 num_layers=1,
                 normalization='layer_norm',
                 norm_eps=1E-5,
                 dropout=0.1,
                 data_dropout=False,
                 activation='leaky',
                 weight_initializer=None,
                 bias_initializer=None,
                 prefix=None, params=None):
        """
        data -> [dropout] * (0/1) -> [Dense -> Normalization -> ACT] * N -> dropout -> Dense -> out

        Parameters
        ----------
        in_units
        mid_units
        out_units
        num_layers
            Number of intermediate layers
        normalization
        norm_eps
        dropout
        activation
        """
        super().__init__(prefix=prefix, params=params)
        self.in_units = in_units
        self.data_dropout = data_dropout
        if mid_units < 0:
            mid_units = in_units
        with self.name_scope():
            self.proj = nn.HybridSequential()
            with self.proj.name_scope():
                if num_layers > 0 and data_dropout:
                    self.proj.add(nn.Dropout(dropout))
                for i in range(num_layers):
                    self.proj.add(nn.Dense(units=mid_units,
                                           in_units=in_units,
                                           flatten=False,
                                           weight_initializer=weight_initializer,
                                           bias_initializer=bias_initializer,
                                           use_bias=False))
                    self.proj.add(get_norm_layer(normalization,
                                                 axis=-1,
                                                 epsilon=norm_eps,
                                                 in_channels=mid_units))
                    self.proj.add(get_activation(activation))
                    in_units = mid_units
                self.proj.add(nn.Dropout(dropout))
                self.proj.add(nn.Dense(units=out_units,
                                       in_units=in_units,
                                       weight_initializer=weight_initializer,
                                       bias_initializer=bias_initializer,
                                       flatten=False))

    def hybrid_forward(self, F, x):
        return self.proj(x)


@use_np
class CategoricalFeatureNet(HybridBlock):
    def __init__(self, num_class, out_units, cfg=None, prefix=None, params=None):
        super().__init__(prefix=prefix, params=params)
        if cfg is None:
            cfg = CategoricalFeatureNet.get_cfg()
        else:
            cfg = CategoricalFeatureNet.get_cfg().clone_merge(cfg)
        self.cfg = cfg
        embed_initializer = mx.init.create(*cfg.initializer.embed)
        weight_initializer = mx.init.create(*cfg.initializer.weight)
        bias_initializer = mx.init.create(*cfg.initializer.bias)
        with self.name_scope():
            self.embedding = nn.Embedding(input_dim=num_class,
                                          output_dim=cfg.emb_units,
                                          weight_initializer=embed_initializer)
            self.proj = BasicMLP(in_units=cfg.emb_units,
                                 mid_units=cfg.mid_units,
                                 out_units=out_units,
                                 num_layers=cfg.num_layers,
                                 normalization=cfg.normalization,
                                 norm_eps=cfg.norm_eps,
                                 data_dropout=cfg.data_dropout,
                                 dropout=cfg.dropout,
                                 activation=cfg.activation,
                                 weight_initializer=weight_initializer,
                                 bias_initializer=bias_initializer)

    @staticmethod
    def get_cfg(key=None):
        if key is None:
            cfg = CfgNode()
            cfg.emb_units = 32
            cfg.mid_units = 64
            cfg.num_layers = 1
            cfg.data_dropout = False
            cfg.dropout = 0.1
            cfg.activation = 'leaky'
            cfg.normalization = 'layer_norm'
            cfg.norm_eps = 1e-5
            cfg.initializer = CfgNode()
            cfg.initializer.embed = ['xavier', 'gaussian', 'in', 1.0]
            cfg.initializer.weight = ['xavier', 'uniform', 'avg', 3.0]
            cfg.initializer.bias = ['zeros']
            return cfg
        else:
            raise NotImplementedError

    def hybrid_forward(self, F, feature):
        embed = self.embedding(feature)
        return self.proj(embed)


@use_np
class NumericalFeatureNet(HybridBlock):
    def __init__(self, input_shape, out_units, cfg=None, prefix=None, params=None):
        super().__init__(prefix=prefix, params=params)
        if cfg is None:
            cfg = NumericalFeatureNet.get_cfg()
        self.input_shape = input_shape
        self.need_first_reshape = isinstance(input_shape, (list, tuple)) and len(input_shape) != 1
        self.in_units = int(np.prod(input_shape))
        self.cfg = NumericalFeatureNet.get_cfg().clone_merge(cfg)
        weight_initializer = mx.init.create(*cfg.initializer.weight)
        bias_initializer = mx.init.create(*cfg.initializer.bias)
        with self.name_scope():
            if self.cfg.input_centering:
                self.data_bn = nn.BatchNorm(in_channels=self.in_units)
            self.proj = BasicMLP(in_units=self.in_units,
                                 mid_units=cfg.mid_units,
                                 out_units=out_units,
                                 num_layers=cfg.num_layers,
                                 normalization=cfg.normalization,
                                 norm_eps=cfg.norm_eps,
                                 data_dropout=cfg.data_dropout,
                                 dropout=cfg.dropout,
                                 activation=cfg.activation,
                                 weight_initializer=weight_initializer,
                                 bias_initializer=bias_initializer)

    @staticmethod
    def get_cfg(key=None):
        if key is None:
            cfg = CfgNode()
            cfg.input_centering = False
            cfg.mid_units = 128
            cfg.num_layers = 1
            cfg.data_dropout = False
            cfg.dropout = 0.1
            cfg.activation = 'leaky'
            cfg.normalization = 'layer_norm'
            cfg.norm_eps = 1e-5
            cfg.initializer = CfgNode()
            cfg.initializer.weight = ['xavier', 'uniform', 'avg', 3.0]
            cfg.initializer.bias = ['zeros']
        else:
            raise NotImplementedError
        return cfg

    def hybrid_forward(self, F, features):
        if self.need_first_reshape:
            features = F.np.reshape(features, (-1, self.in_units))
        if self.cfg.input_centering:
            features = self.data_bn(features)
        return self.proj(features)


@use_np
class FeatureAggregator(HybridBlock):
    def __init__(self, num_fields, out_shape, in_units,
                 cfg=None, prefix=None, params=None):
        super().__init__(prefix=prefix, params=params)
        if cfg is None:
            cfg = FeatureAggregator.get_cfg()
        self.cfg = FeatureAggregator.get_cfg().clone_merge(cfg)
        self.num_fields = num_fields
        if isinstance(out_shape, list):
            out_shape = tuple(out_shape)
        self.out_shape = out_shape
        self.in_units = in_units
        weight_initializer = mx.init.create(*self.cfg.initializer.weight)
        bias_initializer = mx.init.create(*self.cfg.initializer.bias)
        out_units = int(np.prod(out_shape))
        with self.name_scope():
            if self.cfg.agg_type == 'mean':
                in_units = in_units
            elif self.cfg.agg_type == 'concat':
                in_units = in_units * num_fields
            else:
                raise NotImplementedError
            mid_units = in_units if cfg.mid_units < 0 else cfg.mid_units
            self.proj = BasicMLP(in_units=in_units,
                                 mid_units=mid_units,
                                 out_units=out_units,
                                 num_layers=cfg.num_layers,
                                 normalization=cfg.normalization,
                                 norm_eps=cfg.norm_eps,
                                 dropout=cfg.dropout,
                                 data_dropout=cfg.data_dropout,
                                 activation=cfg.activation,
                                 weight_initializer=weight_initializer,
                                 bias_initializer=bias_initializer)

    @staticmethod
    def get_cfg(key=None):
        if key is None:
            cfg = CfgNode()
            cfg.agg_type = 'concat'
            cfg.mid_units = -1
            cfg.num_layers = 0
            cfg.data_dropout = False
            cfg.dropout = 0.1
            cfg.activation = 'tanh'
            cfg.normalization = 'layer_norm'
            cfg.norm_eps = 1e-5
            cfg.initializer = CfgNode()
            cfg.initializer.weight = ['xavier', 'uniform', 'avg', 3.0]
            cfg.initializer.bias = ['zeros']
        else:
            raise NotImplementedError
        return cfg

    def hybrid_forward(self, F, field_proj_features):
        """

        Parameters
        ----------
        field_proj_features
            List of projection features. All elements must have the same shape.

        Returns
        -------
        scores
            Shape (batch_size,) + out_shape
        """
        if len(field_proj_features) == 0:
            agg_features = field_proj_features[0]
        else:
            if self.cfg.agg_type == 'mean':
                agg_features = F.np.stack(field_proj_features)
                agg_features = F.np.mean(agg_features, axis=0)
            elif self.cfg.agg_type == 'concat':
                agg_features = F.np.concatenate(field_proj_features, axis=-1)
            else:
                # TODO(sxjscience) May try to implement more advanced pooling methods for
                #  multimodal data.
                raise NotImplementedError
        scores = self.proj(agg_features)
        if len(self.out_shape) != 1:
            scores = F.np.reshape(scores, (-1,) + self.out_shape)
        return scores


@use_np
class BERTForTabularBasicV1(HybridBlock):
    """The basic model for tabular classification + regression with
    BERT (and its variants like ALBERT, MobileBERT, ELECTRA, etc.)
    as the backbone for handling text data.

    Here, we use the backbone network to extract the contextual embeddings and use
    another dense layer to map the contextual embeddings to the class scores.

    Input:

    TextField + EntityField --> TextNet -------> TextFeature
        ...
    CategoricalField --> CategoricalNet --> CategoricalFeature  ==> AggregateNet --> logits/scores
        ...
    NumericalField ----> NumericalNet ----> NumericalFeature
    """
    def __init__(self, text_backbone,
                 feature_field_info,
                 label_shape=None,
                 cfg=None,
                 prefix=None,
                 params=None):
        """

        Parameters
        ----------
        text_backbone
            Backbone network for handling the text data
        feature_field_info
            The field information of the training data. Each will be a tuple:
            - (field_type, attributes)
        label_shape
            The shape of the label/number of classes. If we need a scalar, it will be an empty tuple "()".
        cfg
            The configuration of the network
        prefix
        params
        """
        super().__init__(prefix=prefix, params=params)
        self.cfg = BERTForTabularBasicV1.get_cfg()
        if cfg is not None:
            self.cfg = self.cfg.clone_merge(cfg)
        assert self.cfg.text_net.pool_type == 'cls'
        feature_units = self.cfg.feature_units
        if feature_units == -1:
            feature_units = text_backbone.units
        if isinstance(label_shape, int):
            out_shape = (label_shape,)
        elif label_shape is None:
            out_shape = ()
        else:
            out_shape = label_shape
        with self.name_scope():
            self.text_backbone = text_backbone
            self.feature_field_info = feature_field_info
            self.categorical_fields = []
            self.numerical_fields = []
            self.agg_layer = FeatureAggregator(num_fields=len(feature_field_info),
                                               out_shape=out_shape,
                                               in_units=feature_units,
                                               cfg=cfg.agg_net)
            self.categorical_networks = None
            self.numerical_network = None
            numerical_elements = None
            for i, (field_type_code, field_attrs) in enumerate(self.feature_field_info):
                if field_type_code == _C.CATEGORICAL:
                    if self.categorical_networks is None:
                        self.categorical_networks = nn.HybridSequential()
                    with self.categorical_networks.name_scope():
                        self.categorical_networks.add(
                            CategoricalFeatureNet(num_class=field_attrs['prop'].num_class,
                                                  out_units=feature_units,
                                                  cfg=cfg.categorical_net))
                elif field_type_code == _C.NUMERICAL:
                    if numerical_elements is None:
                        numerical_elements = int(np.prod(field_attrs['prop'].shape))
                    else:
                        numerical_elements += int(np.prod(field_attrs['prop'].shape))
            if numerical_elements is not None:
                self.numerical_network = NumericalFeatureNet(input_shape=(numerical_elements,),
                                                             out_units=feature_units,
                                                             cfg=cfg.numerical_net)
            else:
                self.numerical_network = None

    @staticmethod
    def get_cfg(key=None):
        if key is None:
            cfg = CfgNode()
            cfg.feature_units = -1  # -1 means not given and we will use the units of BERT
            # TODO(sxjscience) Use a class to store the TextNet
            cfg.text_net = CfgNode()
            cfg.text_net.use_segment_id = True
            cfg.text_net.pool_type = 'cls'
            cfg.agg_net = FeatureAggregator.get_cfg()
            cfg.categorical_net = CategoricalFeatureNet.get_cfg()
            cfg.numerical_net = NumericalFeatureNet.get_cfg()
            cfg.initializer = CfgNode()
            cfg.initializer.weight = ['truncnorm', 0, 0.02]
            cfg.initializer.bias = ['zeros']
            return cfg
        else:
            raise NotImplementedError

    def initialize_with_pretrained_backbone(self, backbone_params_path, ctx=None):
        self.text_backbone.load_parameters(backbone_params_path, ctx=ctx)
        self.agg_layer.initialize(ctx=ctx)
        if self.categorical_networks is not None:
            self.categorical_networks.initialize(ctx=ctx)
        if self.numerical_network is not None:
            self.numerical_network.initialize(ctx=ctx)

    def hybrid_forward(self, F, features):
        """

        Parameters
        ----------
        features
            A list of field data

        Returns
        -------
        logits_or_scores
            Shape (batch_size,) + out_shape
        """
        field_features = []
        text_contextual_features = dict()
        categorical_count = 0
        numerical_samples = []
        for i, (field_type_code, field_attrs) in enumerate(self.feature_field_info):
            if field_type_code == _C.TEXT:
                batch_token_ids, batch_valid_length, batch_segment_ids, _ = features[i]
                if self.cfg.text_net.use_segment_id:
                    contextual_embedding, pooled_output = self.text_backbone(batch_token_ids,
                                                                             batch_segment_ids,
                                                                             batch_valid_length)
                else:
                    contextual_embedding = self.text_backbone(batch_token_ids, batch_valid_length)
                    pooled_output = contextual_embedding[:, 0, :]
                text_contextual_features[i] = contextual_embedding
                field_features.append(pooled_output)
            elif field_type_code == _C.ENTITY:
                # TODO Implement via segment-pool
                raise NotImplementedError('Currently not supported')
            elif field_type_code == _C.CATEGORICAL:
                batch_sample = features[i]
                extracted_feature = self.categorical_networks[categorical_count](batch_sample)
                categorical_count += 1
                field_features.append(extracted_feature)
            elif field_type_code == _C.NUMERICAL:
                batch_sample = features[i]
                numerical_samples.append(F.np.reshape(
                    batch_sample, (-1, int(np.prod(field_attrs['prop'].shape)))))
        if len(numerical_samples) > 0:
            if len(numerical_samples) == 1:
                numerical_feature = self.numerical_network(numerical_samples[0])
            else:
                numerical_feature = self.numerical_network(F.np.concatenate(numerical_samples,
                                                                            axis=-1))
            field_features.append(numerical_feature)
        return self.agg_layer(field_features)
